/** \file
 * Calculate the number of articulation points
 *
 * Accepts a filename on the command line, or
 * simply '-' to read from standard input.
 *
 * http://www.nigels.com/research/
 *
 * \author  Nigel Stewart (nigels@nigels.com)
 * \author  RMIT Dept. Mechanical and Manufacturing Engineering, Australia
 */

#include <csg/model.h>
#include <csg/element.h>
#include <csg/normalise.h>
#include <csg/bbox.h>

#include <csg/scs.h>

#include <iostream>
#include <fstream>
#include <string>
using namespace std;

int main(int argc,char *argv[])
{
    if (argc!=2)
    {
        cout << endl;
        cout << "Articulation Points" << endl;
        cout << "Nigel Stewart (nigels@nigels.com)" << endl;
        cout << "RMIT Department of Mechanical and Manufacturing Engineering" << endl;
        cout << "18th August 2001" << endl;
        cout << endl;
        cout << "Usage: artpoint [filename]" << endl;
        cout << endl;
        cout << "\t filename of input, or '-'" << endl;
        cout << endl;

        return EXIT_FAILURE;
    }

    CsgModel model;

    string filename(argv[1]);

    if (filename=="-")
    {
        if (!model.read(cin))
            return EXIT_FAILURE;
    }
    else
    {
        ifstream is(argv[1]);
        if (!model.read(is))
            return EXIT_FAILURE;
    }

    //
    //
    //

    CsgSumOfProducts sum = normalise(model.tree(0));
    sum.optimise();

    for (uint32 i=0; i<sum.size(); i++)
        if (sum[i].size()>1)
        {
            CsgProduct &prod  = sum[i];
            CsgOverlapGraph   &graph = prod.graph();

            boundingBoxAdjency(graph,sum[i]);

            // Remove intersected nodes

            for (uint32 j=0; j<prod.size(); j++)
                if (prod.positive(j))
                    graph.clear(j);

            // Remove orphans

            CsgOverlapGraph::boolv processed(graph.size(),false);
            CsgOverlapGraph::uintv  nodes;
            graph.orphanNodes(nodes,processed);

            // Remove leaf nodes

            for (;;)
            {
                CsgOverlapGraph::uintv nodes;
                graph.leafNodes(nodes,processed);

                if (nodes.size()==0)
                    break;

                graph.clear(nodes.begin(),nodes.end());
            }

            // Find articulation points
            CsgOverlapGraph::uintv artPoints;
            graph.articulationPoints(artPoints);

            cout << artPoints.size() << endl;
        }

    return EXIT_SUCCESS;
}
