// Experimenting with 2D clipping of convex polygons
//
// Nigel Stewart
// RMIT Dept. Manufacturing and Mechanical Engineering
// nigels@nigels.com
//

#include <vector>
#include <list>
#include <iostream>
using namespace std;

#include <math/vector3.h>
#include <math/plane.h>

#include <cstring>
#include <cassert>
#include <cstdio>

//
//
//

class ConvexPolygons;

//
//

class ConvexPolygon
{
public:

    ConvexPolygon();

    // Construction

    void addPoint(const Vector &pnt);
    void constructionDone();

    //

    bool empty() const;
    bool inside(const Vector &pnt) const;

    //

    friend ConvexPolygon  intersection(const ConvexPolygon &a,const ConvexPolygon &b);
    friend ConvexPolygons difference  (const ConvexPolygon &a,const ConvexPolygon &b);

    friend void clipConvexPolygon
    (
        const ConvexPolygon &poly,
        const Plane         &plane,
              ConvexPolygon &inside,
              ConvexPolygon &outside
    );

    //

    friend ostream &operator<<(ostream &,const ConvexPolygon &);

private:

    vector<Vector>   _point;
    vector<Plane>    _plane;

    int              _done;
};

////////////////////////////////////////////////////////////////

class ConvexPolygons : public list<ConvexPolygon>
{
public:

    bool empty() const;
    bool inside(const Vector &pnt) const;

    //

    friend ostream &operator<<(ostream &,const ConvexPolygons &);
};

bool
ConvexPolygons::empty() const
{
    if (!size())
        return true;

    for (ConvexPolygons::const_iterator i=begin(); i!=end(); i++)
        if (!(*i).empty())
            return false;

    return true;
}

bool
ConvexPolygons::inside(const Vector &pnt) const
{
    for (ConvexPolygons::const_iterator i=begin(); i!=end(); i++)
        if ((*i).inside(pnt))
            return true;

    return false;
}

ostream &
operator<<(ostream &os,const ConvexPolygons &p)
{
    os << p.size() << "ConvexPolygons:" << endl;
    for (ConvexPolygons::const_iterator i=p.begin(); i!=p.end(); i++)
        os << (*i) << endl;

    return os;
}

/////////////////////////////////////////////////////

ConvexPolygon::ConvexPolygon()
: _done(0)
{
    _point.reserve(3);
    _plane.reserve(3);
}

void
ConvexPolygon::addPoint(const Vector &pnt)
{
    assert(!_done);
    _point.push_back(pnt);
}

void
ConvexPolygon::constructionDone()
{
    assert(!_done);
    _done = 1;

    assert(_point.size()>2);

    for (uint32 i=0; i<_point.size(); i++)
    {
        uint32 j = (i+1)%_point.size();

        // Build plane

        Vector normal = xProduct(_point[j] - _point[i],VectorZ);
        Plane  plane(_point[i],normal);

        // Flip the plane normal if necessary

        for (uint32 k=0; k<_point.size(); k++)
        {
            if (k==i || k==j) continue;

            if (plane.inside(_point[k]))
                normal = -normal;
        }

        //

        _plane.push_back(Plane(_point[i],normal));
    }
}

bool
ConvexPolygon::empty() const
{
    return !_point.size();
}

bool
ConvexPolygon::inside(const Vector &pnt) const
{
    assert(_done);

    for (uint32 i=0; i<_plane.size(); i++)
        if (_plane[i].inside(pnt))
            return false;

    return true;
}

ostream &
operator<<(ostream &os,const ConvexPolygon &poly)
{
    os << "ConvexPolygon of " << poly._point.size();
    os << " points and " << poly._plane.size() << " planes." << endl;

    for (uint32 p=0; p<poly._point.size(); p++)
        os << poly._point[p] << endl;

    return os;
}

//
//

ConvexPolygon
intersection(const ConvexPolygon &a,const ConvexPolygon &b)
{
    ConvexPolygon result = a;

    for (uint32 p=0; p<b._plane.size(); p++)
    {
        ConvexPolygon tmp,clipped;
        clipConvexPolygon(result,b._plane[p],tmp,clipped);
        result = clipped;
    }

    return result;
}

ConvexPolygons
difference(const ConvexPolygon &a,const ConvexPolygon &b)
{
    ConvexPolygons result;

    for (uint32 p=0; p<b._plane.size(); p++)
    {
        ConvexPolygon tmp,clipped;
        clipConvexPolygon(a,b._plane[p],clipped,tmp);

        if (!clipped.empty())
            result.push_back(clipped);
    }

    return result;
}

//
//
//

void
clipConvexPolygon
(
    const ConvexPolygon &poly,
    const Plane         &plane,
          ConvexPolygon &inside,
          ConvexPolygon &outside
)
{
    const int points = poly._point.size();

    for (int i=0; i<points; i++)
    {
        int j = (i+1)%points;

        const Vector &p1 = poly._point[i];
        const Vector &p2 = poly._point[j];

        int ic = plane.inside(p1);
        int jc = plane.inside(p2);

        if (ic)
            inside.addPoint(p1);
        else
            outside.addPoint(p1);

        if (ic!=jc)
        {
            Vector intersection = plane.intersectPosition(p1,p2);

            inside.addPoint(intersection);
            outside.addPoint(intersection);
        }
    }

    if (!inside.empty())
        inside.constructionDone();

    if (!outside.empty())
        outside.constructionDone();
}

const int colors = 7;
unsigned char color[colors][3] =
{
    {   0,   0,   0 },
    { 255,   0,   0 },
    {   0, 255,   0 },
    {   0,   0, 255 },
    { 255, 255,   0 },
    { 255,   0, 255 },
    {   0, 255, 255 }
};

int main(int argc,char *argv[])
{
    //

    ConvexPolygon polygon1;

    polygon1.addPoint(Vector(0.1,0.1,0.0));
    polygon1.addPoint(Vector(0.9,0.2,0.0));
    polygon1.addPoint(Vector(0.7,0.9,0.0));
    polygon1.addPoint(Vector(0.1,0.8,0.0));

    polygon1.constructionDone();

    //

    ConvexPolygon polygon2;

    polygon2.addPoint(Vector(0.5,0.2,0.0));
    polygon2.addPoint(Vector(0.9,0.8,0.0));
    polygon2.addPoint(Vector(0.2,0.6,0.0));

    polygon2.constructionDone();

    //

    ConvexPolygon polygon3 = intersection(polygon1,polygon2);

    //

    ConvexPolygons polygon4 = difference(polygon1,polygon2);
    ConvexPolygons polygon5 = difference(polygon2,polygon1);

    //

    cout << "POLYGON4" << endl;
    cout << polygon4 << endl;

    cout << "POLYGON5" << endl;
    cout << polygon5 << endl;

    const int mapRes = 500;
    FILE *fout = fopen("out.ppm","wb");

    unsigned char buffer[1024];
    sprintf((char *) buffer,"P6 %d %d 255\n",mapRes,mapRes);

    fwrite(buffer,1,strlen((char *) buffer),fout);

    for (int y=mapRes-1; y>=0; y--)
        for (int x=0; x<mapRes; x++)
        {
            Vector pos = VectorX*0.002*x+VectorY*0.002*y;
            int c = polygon5.inside(pos);
            fwrite(color[c],1,3,fout);
        }

    fclose(fout);

    return 1;
}
