/** \file
 *  \brief Convert ANCA Volumetric VRML to CSG data
 *
 *   http://www.nigels.com/research/
 *
 *  \author  Nigel Stewart (nigels@nigels.com)
 *  \author  RMIT Dept. Mechanical and Manufacturing Engineering, Australia
 *  \author  RMIT School of Computer Science and Information Technology
 */

#include <glt/rgb.h>

#include <mesh/mesh.h>

#include <node/shapes.h>

#include <string>
#include <vector>
#include <cstdio>
#include <fstream>
#include <iostream>
using namespace std;

void writePoint(ostream &os,const Point &point)
{
    os << point.position().x() << ' ';
    os << point.position().y() << ' ';
    os << point.position().z() << ' ';

    os << point.normal().x() << ' ';
    os << point.normal().y() << ' ';
    os << point.normal().z() << ' ';
}

void writeTriangle(ostream &os,const Triangle &triangle)
{
    writePoint(os,*triangle[0]);
    writePoint(os,*triangle[1]);
    writePoint(os,*triangle[2]);
}

void writeQuad(ostream &os,const Quad &quad)
{
    writePoint(os,*quad[0]);
    writePoint(os,*quad[1]);
    writePoint(os,*quad[2]);
    writePoint(os,*quad[3]);
}

void writeTriangleVolume(ostream &os,const Mesh &mesh,const int i)
{
    const Triangle &tri = mesh.triangle()[i];
    const Point *p[3] = { tri[0], tri[1], tri[2] };
    const Point *o[3] = { NULL,   NULL,   NULL   };

    deque<Point>::const_iterator j;
    for (j=mesh.point().begin(); j!=mesh.point().end(); j++)
    {
        for (int k=0; k<3; k++)
            if (&(*j)==p[k])
            {
                j++;
                o[k] = &(*j);
            }
    }

    assert(o[0]!=NULL);
    assert(o[1]!=NULL);
    assert(o[2]!=NULL);

    if (o[0]!=NULL && o[1]!=NULL && o[2]!=NULL)
    {
        Point *a = (Point *) p[0];
        Point *b = (Point *) p[1];
        Point *c = (Point *) p[2];

        Point *d = (Point *) o[0];
        Point *e = (Point *) o[1];
        Point *f = (Point *) o[2];

        os << "mesh ";

        // 2 triangles
        os << " 2 ";
        writeTriangle(os,Triangle(c,b,a));
        writeTriangle(os,Triangle(d,e,f));

        // 3 quads
        os << " 3 ";
        writeQuad(os,((*a)<(*b) ? Quad(a,b,e,d) : Quad(b,a,d,e)));
        writeQuad(os,((*b)<(*c) ? Quad(b,c,f,e) : Quad(c,b,e,f)));
        writeQuad(os,((*c)<(*a) ? Quad(c,a,d,f) : Quad(a,c,f,d)));

        os << " lightSteelBlue ";
    }
}

int main(int argc,char *argv[])
{
    if (argc<2)
    {
        cout << endl;
        cout << "Volumetic Mesh Conversion" << endl;
        cout << "Nigel Stewart (nigels@nigels.com)" << endl;
        cout << "RMIT Department of Mechanical and Manufacturing Engineering" << endl;
        cout << "RMIT School of Computer Science and Information Technology"  << endl;
        cout << "26th February 2002" << endl;
        cout << endl;
        cout << "Usage: volmesh [filename] [n] [offset]" << endl;
        cout << endl;
        cout << "\t filename is VRML input file" << endl;
        cout << endl;

        return EXIT_FAILURE;
    }

    Mesh mesh(argv[1],MESH_VRML);
    mesh.color() = slateGrey;

    //

    {
        for (deque<Point>::iterator i=mesh.point().begin(); i!=mesh.point().end(); i++)
        {
            i->normal() *= -1.0;
        }
    }

    //

    cout << "X = cylinder grey scale 8 8 48.0 rotate 0 1 0 -90" <<endl;

    vector<string> subtract;

    const int max    = (argc>2 ? atoi(argv[2]) : mesh.triangle().size());
    const int offset = (argc>3 ? atoi(argv[3]) : 0);

    for (int i=0; i<max; i++)
    {
        char name[1024];
        sprintf(name,"T%d",i);

        subtract.push_back(name);
        cout << name << " = ";

        writeTriangleVolume(cout,mesh,i+offset);
        cout << endl;
    }

    if (subtract.size())
    {
        {
            cout << "Tree = (X-";
            for (uint32 i=0; i<subtract.size()-1; i++)
                cout << "(" << subtract[i] << "+";
            cout << subtract.back();
        }

        {
            for (uint32 i=0; i<subtract.size(); i++)
                cout << ')';
            cout << endl;
        }
    }

    return EXIT_SUCCESS;
}
